﻿/// Copyright (c) 2011 Andrey Veselov. All rights reserved.
/// WebSite: http://andrey.moveax.ru 
/// Email: andrey@moveax.ru
/// This source is subject to the Microsoft Public License (Ms-PL).

namespace MVCDemo.Models.Configuration
{
    using System;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.IO;
    using System.Linq;
    using System.Reflection;
    using System.Web.Mvc;
    using System.Xml.Linq;
    using MVCDemo.Models.Extensions;
    using MVCDemo.Models.MetadataProviders;

    public partial class ModelConfigurationProvider
    {
        private class ModelMetadataProviderAdapter : AssociatedMetadataProvider, IModelMetadataProvider
        {
            #region Available properties list

            private readonly string[] _availableProperties = 
            {
                // string
                "DataTypeName",
                "Description",
                "DisplayFormatString",
                "DisplayName",
                "EditFormatString",
                "NullDisplayText",
                "ShortDisplayName",
                "SimpleDisplayText",
                "TemplateHint",
                "Watermark",

                // bool
                "ConvertEmptyStringToNull",
                "HideSurroundingHtml",
                "IsReadOnly",
                "IsRequired",
                "ShowForDisplay",
                "ShowForEdit",

                // IDictionary<string, object>
                "AdditionalValues"
            };

            #endregion

            private readonly ModelConfigurationProvider _cfgProvider;

            public ModelMetadataProviderAdapter(ModelConfigurationProvider cfgProvider)
            {
                if (cfgProvider == null) {
                    throw new ArgumentNullException();
                }

                this._cfgProvider = cfgProvider;

                this.ValidateAvailablePropertiesList(); // called in debug mode only                
            }

            #region IModelMetadataProvider Members

            public bool IsAbleToProvideMetadata(Type containerType)
            {
                return this._cfgProvider.GetModelConfiguraion(containerType) != null;
            }

            #endregion

            #region AssociatedMetadataProvider method override

            protected override ModelMetadata CreateMetadata(
                IEnumerable<System.Attribute> attributes,
                Type containerType,
                Func<object> modelAccessor,
                Type modelType,
                string propertyName)
            {
                var metadata = new ModelMetadata(
                    this, containerType, modelAccessor, modelType, propertyName);

                if (string.IsNullOrEmpty(propertyName)) {
                    return metadata;
                }

                var cfgData = this._cfgProvider.GetModelConfiguraion(containerType);
                if (cfgData == null) {
                    throw new ArgumentException(
                        string.Format("Unknown type: {0}", containerType.FullName));
                }

                // Get the data for the specified property.
                var xmlPropertyElement = cfgData.Source.Elements("property").FirstOrDefault(
                    element => string.Equals(element.Attribute("name").Value, propertyName));

                if (xmlPropertyElement == null) {
                    return metadata;
                }

                // Set the metadata for the property using XML attributes.
                foreach (var element in xmlPropertyElement.Elements()) {
                    if (!string.Equals(element.Name.LocalName, "metadata")) {
                        continue;
                    }

                    string xmlPropertyName = this.GetAndValidatePropertyName(element);

                    if (string.Equals(xmlPropertyName, "AdditionalValues")) {
                        this.SetMetadataAdditionalValue(
                            metadata, element, cfgData.DefaultResourceType);
                    }
                    else {
                        var metadataProperty = typeof(ModelMetadata).GetProperty(xmlPropertyName);
                        object value = Convert.ChangeType(
                            this.GetValueFromElement(element, cfgData.DefaultResourceType),
                            metadataProperty.PropertyType);

                        metadataProperty.SetValue(metadata, value, null);
                    }
                }

                return metadata;
            }

            #endregion

            #region Private methods

            private string GetAndValidatePropertyName(XElement element)
            {
                string propertyName = element.GetValueOrNull("name");
                if (propertyName == null) {
                    throw new InvalidDataException(
                        string.Format("The name was not set. Element: {0}.", element));
                }

                // Validate the property name.
                var nameFound = this._availableProperties.
                    FirstOrDefault(p => string.Equals(p, propertyName));

                if (nameFound == null) {
                    throw new InvalidDataException(
                        string.Format("Unknown or forbidden property: {0}.", propertyName));
                }

                return propertyName;
            }

            private void SetMetadataAdditionalValue(
                ModelMetadata metadata, XElement element, Type defaultResourceType)
            {
                string key = element.GetValueOrNull("key");
                if (key == null) {
                    throw new InvalidDataException(
                        string.Format("The key was not set. Element: {0}.", element));
                }

                // Set the value's type from XML element (or use default System.String).
                string typeName = element.GetValueOrNull("type");
                if (typeName == null) {
                    typeName = "System.String";
                }

                // Get AdditionalValues property and add the new value.
                var addValProperty = typeof(ModelMetadata).GetProperty("AdditionalValues");
                var addVal = (IDictionary<string, object>)addValProperty.GetValue(metadata, null);
                object value = Convert.ChangeType(
                    this.GetValueFromElement(element, defaultResourceType),
                    Type.GetType(typeName));

                addVal.Add(key, value);
            }

            private string GetValueFromElement(XElement element, Type defaultResourceType)
            {
                // Try to get the value.
                string value = element.GetValueOrNull("value");
                if (value != null) {
                    return value;
                }

                // Try to get the value from the resource.
                string resourceName = element.GetValueOrNull("resname");
                if (resourceName == null) {
                    throw new InvalidDataException(
                        string.Format("The value was not set. Element: {0}.", element));
                }

                string resourceTypeName = element.GetValueOrNull("resource");
                Type resourceType = defaultResourceType;

                if (resourceTypeName != null) {
                    resourceType = Type.GetType(resourceTypeName, throwOnError: true);
                }

                var resourceProperty = resourceType.GetProperty(
                    resourceName,
                    BindingFlags.Static | BindingFlags.Public | BindingFlags.NonPublic);

                return (string)resourceProperty.GetValue(null, null);
            }

            #endregion

            #region Debug methods

            [Conditional("DEBUG")]
            private void ValidateAvailablePropertiesList()
            {
                PropertyInfo[] props = typeof(ModelMetadata).GetProperties();

                foreach (var propertyName in this._availableProperties) {
                    if (props.FirstOrDefault(p => string.Equals(p.Name, propertyName)) == null) {
                        throw new InvalidDataException(
                            string.Format("Unknown ModelMetadata property: {0}.", propertyName));
                    }
                }
            }

            #endregion
        }
    }
}